//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Loaders/QREDataLoaderProperties.cpp
//! @brief     Implements class AutomaticMultiColumnDataLoader1DProperties
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Loaders/QREDataLoaderProperties.h"
#include "ui_QREDataLoaderProperties.h"
#include <QCheckBox>
#include <QtGui>

QREDataLoaderProperties::QREDataLoaderProperties()
    : m_allowFactors(false)
{
    m_ui = new Ui::QREDataLoaderProperties;
    m_ui->setupUi(this);

    m_gridLayout = m_ui->gridLayout;
    m_enableErrorCheckBox = m_ui->enableErrorCheckBox;
    m_headerPrefixEdit = m_ui->headerPrefixEdit;
    m_linesToSkipEdit = m_ui->linesToSkipEdit;
    m_separatorCombo = m_ui->separatorCombo;
    m_qUnitCombo = m_ui->qUnitCombo;

    allowFactors(false);

    connect(m_headerPrefixEdit, &QLineEdit::textChanged, [this]() { emit propertiesChanged(); });
    connect(m_linesToSkipEdit, &QLineEdit::textChanged, [this]() { emit propertiesChanged(); });
    connect(m_separatorCombo, &QComboBox::currentTextChanged,
            [this]() { emit propertiesChanged(); });

    connect(m_enableErrorCheckBox, &QCheckBox::stateChanged, this,
            &QREDataLoaderProperties::onErrorEnablingChanged);

    connect(m_qUnitCombo, QOverload<int>::of(&QComboBox::currentIndexChanged),
            [this]() { emit propertiesChanged(); });

    for (int i = 0; i < 3; i++) {
        // In the following: disable the checkbox before signaling propertyChanged to suppress ghost
        // value changes in case the propertiesChanged signals triggers long calculations (the
        // spinbox would see a timeout in its "pressed"-loop and generate a second value change)
        connect(columnSpinBox(i), QOverload<int>::of(&QSpinBox::valueChanged), [this, i]() {
            QSpinBox* sb = columnSpinBox(i);
            if (sb)
                sb->setEnabled(false);
            emit propertiesChanged();
            if (sb)
                sb->setEnabled(true);
        });

        connect(factorSpinBox(i), QOverload<double>::of(&QDoubleSpinBox::valueChanged),
                [this, i]() {
                    QDoubleSpinBox* sb = factorSpinBox(i);
                    if (sb)
                        sb->setEnabled(false);
                    emit propertiesChanged();
                    if (sb)
                        sb->setEnabled(true);
                });
    }
}

void QREDataLoaderProperties::allowFactors(bool b)
{
    m_allowFactors = b;
    for (int dataType = 0; dataType < 3; dataType++) {
        if (auto* label = factorLabel(dataType))
            label->setVisible(b);
        if (auto* sb = factorSpinBox(dataType))
            sb->setVisible(b);
    }

    updateErrorEnabling(m_enableErrorCheckBox->isChecked());
}

double QREDataLoaderProperties::factor(int dataType) const
{
    auto* const sb = factorSpinBox(dataType);
    return sb && sb->isVisible() ? sb->value() : 1.0;
}

void QREDataLoaderProperties::updateErrorEnabling(bool enabled) const
{
    const int lineInLayout = 2;

    for (int col = 2; col < m_gridLayout->columnCount(); col++) {
        auto* const layoutItem = m_gridLayout->itemAtPosition(lineInLayout, col);
        if (layoutItem) {
            QWidget* w = layoutItem->widget();
            if (w) {
                const bool belongsToFactor = col == factorColumn || col == factorLabelColumn;
                w->setVisible(enabled && (!belongsToFactor || m_allowFactors));
            }
        }
    }
}

void QREDataLoaderProperties::onErrorEnablingChanged()
{
    const bool isEnabled = m_enableErrorCheckBox->isChecked();

    updateErrorEnabling(isEnabled);

    emit propertiesChanged();
}

QSpinBox* QREDataLoaderProperties::columnSpinBox(int dataType) const
{
    const int lineInLayout = dataType;
    QLayoutItem* item = m_gridLayout->itemAtPosition(lineInLayout, columnColumn);
    if (!item)
        return {};
    return dynamic_cast<QSpinBox*>(item->widget());
}

QDoubleSpinBox* QREDataLoaderProperties::factorSpinBox(int dataType) const
{
    const int lineInLayout = dataType;
    QLayoutItem* item = m_gridLayout->itemAtPosition(lineInLayout, factorColumn);
    if (!item)
        return {};
    return dynamic_cast<QDoubleSpinBox*>(item->widget());
}

QLabel* QREDataLoaderProperties::factorLabel(int dataType) const
{
    const int lineInLayout = dataType;
    QLayoutItem* item = m_gridLayout->itemAtPosition(lineInLayout, factorLabelColumn);
    if (!item)
        return {};
    return dynamic_cast<QLabel*>(item->widget());
}
